// Copyright Isaac Z. Schlueter and Contributors. All rights reserved. ISC license.
// Copyright 2018-2023 the Deno authors. All rights reserved. MIT license.
import { assert } from "../assert/mod.ts";
import { parse } from "./parse.ts";
import { parseRange } from "./parse_range.ts";
import { ltr } from "./ltr.ts";

Deno.test("ltr", async (t) => {
  // [range, version]
  // Version should be greater than range
  const versions: [string, string][] = [
    ["~1.2.2", "1.2.1"],
    ["~0.6.1-1", "0.6.1-0"],
    ["1.0.0 - 2.0.0", "0.0.1"],
    ["1.0.0-beta.2", "1.0.0-beta.1"],
    ["1.0.0", "0.0.0"],
    [">=2.0.0", "1.1.1"],
    [">=2.0.0", "1.2.9"],
    [">2.0.0", "2.0.0"],
    ["0.1.20 || 1.2.4", "0.1.5"],
    ["2.x.x", "1.0.0"],
    ["1.2.x", "1.1.0"],
    ["1.2.x || 2.x", "1.0.0"],
    ["2.*.*", "1.0.1"],
    ["1.2.*", "1.1.3"],
    ["1.2.* || 2.*", "1.1.9999"],
    ["2", "1.0.0"],
    ["2.3", "2.2.2"],
    ["~2.4", "2.3.0"], // >=2.4.0 <2.5.0
    ["~2.4", "2.3.5"],
    ["~>3.2.1", "3.2.0"], // >=3.2.1 <3.3.0
    ["~1", "0.2.3"], // >=1.0.0 <2.0.0
    ["~>1", "0.2.4"],
    ["~1.0", "0.1.2"], // >=1.0.0 <1.1.0
    [">1.2", "1.2.0"],
    ["~v0.5.4-pre", "0.5.4-alpha"],
    ["~v0.5.4-pre", "0.5.4-alpha"],
    ["=0.7.x", "0.6.0"],
    ["=0.7.x", "0.6.0-asdf"],
    [">=0.7.x", "0.6.0"],
    ["~1.2.2", "1.2.1"],
    ["1.0.0 - 2.0.0", "0.2.3"],
    ["1.0.0", "0.0.1"],
    [">=2.0.0", "1.0.0"],
    [">=2.0.0", "1.9999.9999"],
    [">=2.0.0", "1.2.9"],
    [">2.0.0", "2.0.0"],
    [">2.0.0", "1.2.9"],
    ["2.x.x", "1.1.3"],
    ["1.2.x", "1.1.3"],
    ["1.2.x || 2.x", "1.1.3"],
    ["2.*.*", "1.1.3"],
    ["1.2.*", "1.1.3"],
    ["1.2.* || 2.*", "1.1.3"],
    ["2", "1.9999.9999"],
    ["2.3", "2.2.1"],
    ["~2.4", "2.3.0"], // >=2.4.0 <2.5.0
    ["~>3.2.1", "2.3.2"], // >=3.2.1 <3.3.0
    ["~1", "0.2.3"], // >=1.0.0 <2.0.0
    ["~>1", "0.2.3"],
    ["~1.0", "0.0.0"], // >=1.0.0 <1.1.0
    [">1", "1.0.0"],
    ["=0.7.x", "0.6.2"],
    ["=0.7.x", "0.7.0-asdf"],
    ["^1", "1.0.0-0"],
    [">=0.7.x", "0.7.0-asdf"],
    [">=0.7.x", "0.6.2"],
    [">1.2.3", "1.2.0-alpha"],
  ];

  for (const [a, b] of versions) {
    await t.step(`${b} < ${a}`, () => {
      const range = parseRange(a);
      const version = parse(b);
      assert(ltr(version, range));
    });
  }
});

Deno.test("ltrNegative", async (t) => {
  // [range, version]
  // Version should be greater than range
  const versions: [string, string][] = [
    ["~0.6.1-1", "0.6.1-1"],
    ["1.0.0 - 2.0.0", "1.2.3"],
    ["1.0.0 - 2.0.0", "2.9.9"],
    ["1.0.0", "1.0.0"],
    [">=*", "0.2.4"],
    ["*", "1.2.3"],
    [">=1.0.0", "1.0.0"],
    [">=1.0.0", "1.0.1"],
    [">=1.0.0", "1.1.0"],
    [">1.0.0", "1.0.1"],
    [">1.0.0", "1.1.0"],
    ["<=2.0.0", "2.0.0"],
    ["<=2.0.0", "1.9999.9999"],
    ["<=2.0.0", "0.2.9"],
    ["<2.0.0", "1.9999.9999"],
    ["<2.0.0", "0.2.9"],
    [">=0.1.97", "v0.1.97"],
    [">=0.1.97", "0.1.97"],
    ["0.1.20 || 1.2.4", "1.2.4"],
    ["0.1.20 || >1.2.4", "1.2.4"],
    ["0.1.20 || 1.2.4", "1.2.3"],
    ["0.1.20 || 1.2.4", "0.1.20"],
    [">=0.2.3 || <0.0.1", "0.0.0"],
    [">=0.2.3 || <0.0.1", "0.2.3"],
    [">=0.2.3 || <0.0.1", "0.2.4"],
    ["||", "1.3.4"],
    ["2.x.x", "2.1.3"],
    ["1.2.x", "1.2.3"],
    ["1.2.x || 2.x", "2.1.3"],
    ["1.2.x || 2.x", "1.2.3"],
    ["x", "1.2.3"],
    ["2.*.*", "2.1.3"],
    ["1.2.*", "1.2.3"],
    ["1.2.* || 2.*", "2.1.3"],
    ["1.2.* || 2.*", "1.2.3"],
    ["1.2.* || 2.*", "1.2.3"],
    ["*", "1.2.3"],
    ["2", "2.1.2"],
    ["2.3", "2.3.1"],
    ["~2.4", "2.4.0"], // >=2.4.0 <2.5.0
    ["~2.4", "2.4.5"],
    ["~>3.2.1", "3.2.2"], // >=3.2.1 <3.3.0
    ["~1", "1.2.3"], // >=1.0.0 <2.0.0
    ["~>1", "1.2.3"],
    ["~1.0", "1.0.2"], // >=1.0.0 <1.1.0
    [">=1", "1.0.0"],
    ["<1.2", "1.1.1"],
    ["~v0.5.4-pre", "0.5.5"],
    ["~v0.5.4-pre", "0.5.4"],
    ["=0.7.x", "0.7.2"],
    [">=0.7.x", "0.7.2"],
    ["<=0.7.x", "0.6.2"],
    [">0.2.3 >0.2.4 <=0.2.5", "0.2.5"],
    [">=0.2.3 <=0.2.4", "0.2.4"],
    ["1.0.0 - 2.0.0", "2.0.0"],
    ["^3.0.0", "4.0.0"],
    ["^1.0.0 || ~2.0.1", "2.0.0"],
    ["^0.1.0 || ~3.0.1 || 5.0.0", "3.2.0"],
    ["^0.1.0 || ~3.0.1 || >4 <=5.0.0", "3.5.0"],
    ["^1.0.0-alpha", "1.0.0-beta"],
    ["~1.0.0-alpha", "1.0.0-beta"],
    ["=0.1.0", "0.1.0"],
  ];

  for (const [a, b] of versions) {
    await t.step(`${b} ≮ ${a}`, () => {
      const range = parseRange(a);
      const version = parse(b);
      assert(!ltr(version, range));
    });
  }
});
